import os
#field which is used to store meta information about the file in ElasticSearch, e.g. path, extension, type
META_FIELD = "properties"
ES_URL = os.getenv("DOCPARSER_ES_URL", "http://elastic-dev.texta.ee:9200")



import logging
from typing import List

from texta_entity_linker.entity_linker import EntityLinker


logging.basicConfig(
    format='%(levelname)s %(asctime)s: %(message)s',
    datefmt='%d.%m.%Y %H:%M:%S',
    level=logging.INFO
)


class EntityLinkerWrapper:
    """
    Takes the MLP generator as input, reads all the content and
    takes the whole input to produce detailed person information
    from the emails content.
    """


    def __init__(self, abbreviations_filepath: str = None):
        self.abbreviations_file = abbreviations_filepath
        self.concatenator = EntityLinker(self.abbreviations_file) if abbreviations_filepath else EntityLinker()


    def concat_from_generator(self, generator):
        """
        Takes a generator as an input which contains tuples containing the MLP
        and attachment information, parses their Texta Facts and then does entity-linking
        on the whole set.
        :param generator: Generator which yields a two-element tuple with dictionaries.
        :return: Yields all the linked information of the entities.
        """
        container = []
        for mlp_output, attachments in generator:
            container.append(mlp_output)

            for attachment in attachments:
                container.append(attachment)

        self.concatenator.from_json(container)

        logging.info("Starting the entity linking process, this might take some time...")
        self.concatenator.link_entities()

        for person_info in self.concatenator.to_json():
            if isinstance(person_info, dict):
                yield person_info
            elif isinstance(person_info, list):
                for item in person_info:
                    yield item


    def concat_from_list(self, mlp_output: List[dict]):
        self.concatenator.from_json(mlp_output)
        self.concatenator.link_entities()
        return self.concatenator.to_json()



from texta_parsers.email_parser import EmailParser
from texta_parsers import exceptions

import mimetypes
import magic
import os



class Extension:

    AUDIO_EXTENSIONS = ()
    DIGIDOC_EXTENSIONS = (".ddoc", ".bdoc", ".asice", ".asics")
    DOC_EXTENSIONS = (".doc", ".docx", ".odt", ".pdf", ".rtf", ".htm", ".html", ".epub", ".txt")
    IMAGE_EXTENSIONS = (".jpg", ".jpeg", ".gif", ".png", ".bmp", ".tiff")
    ARCHIVE_EXTENSIONS = (".zip", ".tar", ".bz2", ".gz") + DIGIDOC_EXTENSIONS
    COLLECTION_EXTENSIONS = (".csv", ".xls", ".xlsx")
    EMAIL_EXTENSIONS = (".mbox", ".pst", ".eml")
    #requires patool to be installed
    #ARCHIVE_EXTENSIONS = (".7z", ".bz2", ".gz", ".lrz", ".lha", ".lzh", ".lz", ".rar", ".rz", ".tar", ".xz", ".zip") + DIGIDOC_EXTENSIONS
    # union of all known extensions used in parsing
    KNOWN_EXTENSIONS = ARCHIVE_EXTENSIONS + DOC_EXTENSIONS + AUDIO_EXTENSIONS + COLLECTION_EXTENSIONS + IMAGE_EXTENSIONS + EMAIL_EXTENSIONS
    # these files are always identified by extension, because magic misidentifies them or because it is particularly important that these 
    # files are identified correctly and we trust extension more than magic for these file types
    EXTENSION_OVERWRITES = DIGIDOC_EXTENSIONS + AUDIO_EXTENSIONS + EMAIL_EXTENSIONS + COLLECTION_EXTENSIONS


    def __init__(self):
        pass


    @staticmethod
    def is_supported(file, file_name=None):
        """
        Tests if file is supported.
        :return: bool.
        """
        try:
            if Extension.predict(file, file_name=file_name) in Extension.KNOWN_EXTENSIONS:
                return True
        except exceptions.UnsupportedFileError:
            return False
        return False


    @staticmethod
    def predict(file, file_name=None):
        """
        Tries to validate magic output with some trivial hacks.
        Magic tends to mess up some files for some reason. Magic also fails on digidoc files.
        So let's do some double checks to be sure we parse it correctly.
        :param: file: either bytes or string to filepath.
        :param: file_name: if file is bytes then file_name is always given. 
        :return: File extension as string.
        """
        if isinstance(file, bytes):
            # input is file as bytes
            mime_type = magic.from_buffer(file, mime=True)
        else:
            # input is a path to file
            mime_type = magic.from_file(file, mime=True)
            file_name = file
        # let magic guess the extension based on mime
        guessed_extensions = mimetypes.guess_all_extensions(mime_type)
        # try extracting extenion from file name
        _, ext = os.path.splitext(file_name)
        parsed_ext = ext.lower()
        # extra check for mail or mailbox files
        file_beginning = EmailParser._try_get_file_beginning(file)
        mail_ext = EmailParser._file_is_mail(file_name, file_beginning)
        # final call
        if mail_ext:
            predicted_ext = mail_ext
        elif parsed_ext in Extension.EXTENSION_OVERWRITES or not guessed_extensions:
            predicted_ext = parsed_ext
        else:
            ### From the list of predicted possible extensions choose the one which is known, if possible
            guessed_known_exts = [ext for ext in guessed_extensions if ext in Extension.KNOWN_EXTENSIONS]
            predicted_ext = guessed_known_exts[0] if len(guessed_known_exts) > 0 else guessed_extensions[0]

        if(predicted_ext not in Extension.KNOWN_EXTENSIONS):
            raise exceptions.UnsupportedFileError("File type not supported.")
        return predicted_ext



import logging
import mailbox
import multiprocessing
import os
import shutil
from random import getrandbits
from typing import Optional

import magic

from texta_parsers.tools.message import AttachmentParser, BodyParser, HeaderParser
from texta_parsers.tools.pst_extractor import check_readpst_exists, extract_pst_file


SUPPORTED_MAIL_FORMATS = [".mbox", ".pst", ".eml"]
logging.basicConfig(format='%(levelname)s %(asctime)s: %(message)s',
                    datefmt='%d.%m.%Y %H:%M:%S',
                    level=logging.INFO)


class EmailParser:
    """ 
    # Parameters 
    max_attachment_size (int): in megabytes. Content from larger attachments won't be parsed (optional, default: 10)
    num_threads (int): number of threads in case of multiprocessing, (optional, default: 1)
    tmp_folder (string): folder for temporal files, (optional, default: "tmp")
    """


    def __init__(self, max_attachment_size=10, num_threads=1, tmp_folder="tmp"):
        self.max_attachment_size = max_attachment_size
        self.num_threads = num_threads
        self.tmp_folder = tmp_folder
        self.max_attachment_size = max_attachment_size
        if not check_readpst_exists():
            raise FileNotFoundError("Readpst not found. Can not initialize parser.")


    def parse(self, input_path: str, attachments_folder: str = None, mails_folder: Optional[str] = None, random_filenames: bool = False):
        """ Parses emails from given folder or file.

        # Parameters
        input_path (str): Path to folder or file that contains emails
        attachments_folder (str): folder where attachments are saved (optional, default: None)
        mails_folder (str): folder where emails are saved as eml files (optional, default: None)
        random_filenames (bool): specifies whether random filenames should be generated for attachments.
        This is only needed if you plan to host files in toolkit's media folder in order to avoid encoding errors in URLs (optional, default: False)
        use_mlp (bool): specifies whether MLP should be applied to parser output (optional, default: False)

        # Returns
        Generator which yields emails in format (msg_json, [attachment_json])
        """

        email_files = self._get_files(input_path)
        self.attachments_folder = attachments_folder
        self.mails_folder = mails_folder
        self.random_filenames = random_filenames
        self._make_dirs()

        for email_file, ext in email_files:
            logging.info("Parsing {}".format(email_file))

            if ext == ".mbox":
                folder_name = email_file.split("/")[-1].split(".")[0]

                with multiprocessing.Pool(processes=self.num_threads) as pool:
                    for msg_json, attachment_jsons in pool.imap(self._parse_msg, mailbox.mbox(email_file)):
                        if (msg_json != {}):
                            msg_json["original_folder"] = folder_name
                        yield msg_json, attachment_jsons

            elif ext == ".pst":
                extracted_mboxes = extract_pst_file(email_file, self.tmp_folder)
                for mbox_file in extracted_mboxes:
                    logging.info("Parsing from .pst {}".format(mbox_file))
                    folder_name = mbox_file.split("/")[-1].split(".")[0]

                    with multiprocessing.Pool(processes=self.num_threads) as pool:
                        for msg_json, attachment_jsons in pool.imap(self._parse_msg, mailbox.mbox(mbox_file)):
                            if (msg_json != {}):
                                msg_json["original_folder"] = folder_name
                            yield msg_json, attachment_jsons

                self._delete_extracted_pst(self.tmp_folder)

            elif ext == ".eml":
                raw_msg = self._load_raw_mail(email_file)
                msg = mailbox.Message(raw_msg)
                msg_json, attachment_jsons = self._parse_msg(msg)
                yield msg_json, attachment_jsons


    """
    Parses given message recursively using parsers in message module for different parts
    """


    def _parse_msg_rec(self, msg, save_path, msg_id):
        is_root_msg = msg_id == ""

        if is_root_msg:  # is main email, parse it accordingly
            msg_data, attachments = self._get_all_msg_data(msg, save_path)
            msg_id = msg_data["msg_id"]

        else:  # is message/rfc822 attachment, parse it as an email with respective parsers but return as an attachment
            try:
                filename = AttachmentParser.parse_filename(msg.get_filename())
            except:
                filename = ""

            msg = msg.get_payload()[0]
            msg_data, attachments = self._get_all_msg_data(msg, save_path)

            attachment_data = {}
            attachment_data["msg_id"] = msg_id
            attachment_data["content"] = msg_data["body"]
            attachment_data["attachment_id"] = msg_data["msg_id"]
            attachment_data["filename"] = filename
            if "errors" in msg_data and "body" in msg_data["errors"]:
                attachment_data["errors"] = {"content": msg_data["errors"]["body"]}

            for attachment in attachments:
                attachment["parent_id"] = msg_data["msg_id"]
                attachment["msg_id"] = msg_id  # use existing msg_id

            # add parsed email header and body also to attachments
            attachments.append(attachment_data)

        # parse other message/rfc822 attachments if there is any
        if msg.is_multipart():
            for payload in msg.get_payload():
                if payload.get_content_type() == "message/rfc822":
                    _, nested_attachments = self._parse_msg_rec(payload, save_path, msg_id)
                    attachments += nested_attachments

        return msg_data, attachments


    def _parse_msg(self, msg):
        try:
            msg_json, attachment_jsons = self._parse_msg_rec(msg, self.attachments_folder, "")
            if (self.mails_folder):
                filepath = self._save_mail_as_eml(msg, msg_json["msg_id"])
                if (filepath):
                    msg_json["filepath"] = filepath
            return msg_json, attachment_jsons
        except Exception as e:
            # in case all other more specific exceptions handlers have failed
            logging.exception(e)
            return {}, [{}]


    def _get_all_msg_data(self, msg, save_path, ):
        h_parser = HeaderParser(msg)
        b_parser = BodyParser(msg)

        msg_data = h_parser.parse()
        body = b_parser.parse_body()
        msg_data["body"] = body

        msg_id = str(getrandbits(64))
        msg_data["msg_id"] = msg_id

        attachments = b_parser.parse_attachments(save_path, msg_id, self.max_attachment_size, self.random_filenames)

        # concatenate errors
        header_fails = h_parser.fails
        header_fails.update(b_parser.fails)

        if header_fails != {}:
            msg_data["errors"] = header_fails

        return msg_data, attachments


    @staticmethod
    def _try_get_file_beginning(file):
        if isinstance(file, bytes):
            return file[:2048]
        else:
            try:
                return open(file, "rb").read(2048)
            except:
                return None


    def _get_files(self, path):
        """Returns a list of email files from given folder (recursive) or a list with single file if input is file.
        """
        results = []
        if os.path.isfile(path):
            file_beginning = EmailParser._try_get_file_beginning(path)
            res = self._file_is_mail(path, file_beginning)
            if (res):
                results.append((path, res))
        else:
            for dirpath, _, filenames in os.walk(path):
                for filename in filenames:
                    full_file = os.path.join(dirpath, filename)

                    file_beginning = EmailParser._try_get_file_beginning(full_file)
                    res = self._file_is_mail(filename, file_beginning)
                    if (res):
                        results.append((full_file, res))

        return results


    @staticmethod
    def _file_is_mail(filename, file_beginning):
        ext = os.path.splitext(filename)[1]
        if ext in SUPPORTED_MAIL_FORMATS:
            return ext

        if (file_beginning == None):
            return
        elif (EmailParser._has_msg_mimetype(file_beginning)):
            return ".eml"
        elif (EmailParser._begins_like_mbox(file_beginning)):
            return ".mbox"


    @staticmethod
    def _has_msg_mimetype(input):
        mimetype = magic.from_buffer(input, mime=True)
        return mimetype == "message/rfc822"


    @staticmethod
    def _begins_like_mbox(input):
        try:
            return input.decode().startswith("From ")
        except:
            return False


    def _load_raw_mail(self, path):
        with open(path, "rb") as f:
            raw_mail = f.read()
        return raw_mail


    def _make_dirs(self):
        for path in [self.tmp_folder, self.mails_folder, self.attachments_folder]:
            if (path != None):
                if not os.path.exists(path):
                    os.makedirs(path)


    def _save_mail_as_eml(self, message, msg_id):
        msg_path = os.path.join(self.mails_folder, msg_id + ".eml")
        try:
            with open(msg_path, "wb") as f:
                f.write(bytes(message))
            return msg_path
        except KeyError:
            logging.error("Failed flattening file with msg_id {}".format(msg_id))
            return None


    def _delete_extracted_pst(self, pst_dir):
        """Deletes the content of the tmp folder which is used for temporarily extracting pst files
        """
        for root, dirs, files in os.walk(pst_dir):
            for f in files:
                os.unlink(os.path.join(root, f))
            for d in dirs:
                shutil.rmtree(os.path.join(root, d))



import logging

from . import utils
from .utils import ParserOutputType


logging.basicConfig(
    format='%(levelname)s %(asctime)s: %(message)s',
    datefmt='%d.%m.%Y %H:%M:%S',
    level=logging.INFO
)


class MLPProcessor:

    def __init__(self, mlp):
        self.mlp = mlp


    def apply_mlp(self, generator):
        """ Applies MLP to objects in a given generator.
        """
        for item in generator:
            item_type = utils.get_output_type(item)

            if (item_type == ParserOutputType.EMAIL):
                email, attachments = item
                self._apply_mlp_to_mails(email, attachments)

            elif (item_type == ParserOutputType.COLLECTION):
                self._apply_mlp_to_collection_item(item)
            else:
                self._apply_mlp(item, "text")

            yield item


    def _apply_mlp(self, document: dict, field: str):
        if (field not in document):
            return
        content = document.get(field, "")

        if content:
            mlp_res = self.mlp.process(content)
            mlp_res_path = field + "_mlp"
            document[mlp_res_path] = mlp_res["text"]

            facts = []
            for f in mlp_res["texta_facts"]:
                f["doc_path"] = [mlp_res_path + "." + f["doc_path"]]
                facts.append(f)

            if facts:
                document["texta_facts"] = facts


    def _apply_mlp_to_mails(self, email, attachments):
        self._apply_mlp(email, "body")
        for attachment in attachments:
            self._apply_mlp(attachment, "content")


    # apply it to all fields as we don't know anything about the item or it's fields
    def _apply_mlp_to_collection_item(self, item: dict):
        item_copy = item.copy()
        for key in item_copy.keys():
            self._apply_mlp(item, key)



import subprocess
import os
import logging

def _extract_pst_file(pst_filepath, tmp_dir):
    """ Extracts emails and attachments from .pst file located in pst_filepath
        to separate files located in outdir.
    """
    logging.info('Converting {}...'.format(pst_filepath))
    return subprocess.run(["readpst", pst_filepath, "-o", tmp_dir])

def _get_mbox_files(tmp_dir):
    """ Returns a list of full paths to mbox files.
    """
    result = []
    for dirpath, _, filenames in os.walk(tmp_dir):
        for filename in filenames:
            _, ext = os.path.splitext(filename)
            if ext == ".mbox":
                result.append(os.path.join(dirpath, filename))
    return result

def extract_pst_file(pst_file, tmp_dir):
    """ Extracts emails and attachments from given .pst file
        to separate files (and subdirectories) located in outdir.

        Returns full path from base directory to extracted mbox files.
    """
    _extract_pst_file(pst_file, tmp_dir)

    return _get_mbox_files(tmp_dir)


def check_readpst_exists():
    try:
        p = subprocess.run(["readpst"], stdout=subprocess.PIPE)
        return True
    except FileNotFoundError:
        return False



import os

from texta_parsers.tools.extension import Extension


class DocScanner:

    def __init__(self):
        self.scanned_files_path = os.path.join(os.getcwd(), "scanned_files.txt")
        self.scanned_files = self._get_scanned_files()


    def scan(self, directory):
        """
        Scans directory and all its subdirectories.
        All paths for found (and supported) files are yielded.
        """
        batch = []
        for dirpath, _, filenames in os.walk(directory):
            for filename in filenames:
                file_path = os.path.join(dirpath, filename)
                abspath = os.path.abspath(file_path)
                print(abspath)
                if abspath not in self.scanned_files:
                    self._save_path(abspath)
                    print(abspath)
                    if Extension.is_supported(abspath):
                        yield abspath


    def _save_path(self, path):
        """
        Saves seen paths to file.
        """
        with open(self.scanned_files_path, "a") as f:
            self.scanned_files[path] = True
            f.write(f"{path}\n")


    def _get_scanned_files(self):
        """
        Retrieves saved paths fron file.
        """
        if(os.path.isfile(self.scanned_files_path) == True):
            with open(self.scanned_files_path, "r") as f:
                files_dict = {path.strip(): True for path in f.readlines() if path}
            return files_dict
        else:
            return {}


    def remove_scanned_file(self):
        """
        Deletes file containing scanned paths.
        """
        if os.path.isfile(self.scanned_files_path):
            os.remove(self.scanned_files_path)



from bs4 import BeautifulSoup
from random import getrandbits
import email
import sys
from tika import parser as tika_parser
import os
import uuid
import logging


DIGIDOC_FORMATS = ['.ddoc', '.bdoc', '.asice', '.asics']


class HeaderParser:
    def __init__(self, msg):
        self.msg = msg
        self.fails = {}

    def parse(self):
        header_data = {}
        header_data["from"] = self._get_addr("from")
        header_data["to"] = self._get_addr("to")
        header_data["subject"] = self._get_subject()
        header_data["date"] = self._get_date()

        cc = self._get_addr("cc")
        bcc = self._get_addr("bcc")

        if(cc):
            header_data["cc"] = cc
        if(bcc):
            header_data["bcc"] = bcc

        return header_data

    def _get_addr(self, field):
        try:
            addr_list = email.utils.getaddresses(self.msg.get_all(field, []))

            all_addresses = []
            for addr_tuple in addr_list:
                addr_items = []
                for item in addr_tuple:
                    item_dec, charset = email.header.decode_header(item)[0]

                    if charset:
                        item_str = item_dec.decode(charset)
                    elif type(item_dec) == bytes:
                        item_str = item_dec.decode()
                    else:
                        assert type(item) == str
                        item_str = item_dec

                    if(item_str):
                        addr_items.append(item_str)
                all_addresses.append(" ".join(addr_items))

            return "\n".join(all_addresses)

        except:
            exc_type, value, _ = sys.exc_info()
            self.fails[field] = "{}: {}".format(exc_type.__name__, value)
            return ""

    def _get_subject(self):
        if "subject" in self.msg:
            try:
                subject_parts = email.header.decode_header(self.msg["subject"])

                subject_items = []
                for item_dec, charset in subject_parts:
                    if charset:
                        item_str = item_dec.decode(charset)
                    # charset is None for non-encoded header parts and the part itself a bytestring
                    elif type(item_dec) == bytes:
                        item_str = item_dec.decode()
                    else:
                        # if whole header is non-encoded, result is a string
                        assert type(item_dec) == str
                        item_str = item_dec

                    subject_items.append(item_str.strip())

                return " ".join(subject_items)

            except:
                exc_type, value, _ = sys.exc_info()
                self.fails["subject"] = "{}: {}".format(exc_type.__name__, value)
                return ""
        else:
            return ""

    def _get_date(self):
        if "date" in self.msg:
            date = self.msg["date"]
            if(type(date) == str):
                return email.utils.parsedate_to_datetime(self.msg["date"])
            elif(type(date) == email.header.Header):
                encoded_date, charset = email.header.decode_header(self.msg["date"])[0]
                decoded_date = None
                try:
                    decoded_date = encoded_date.decode(charset)
                except:
                    decoded_date = encoded_date.decode("Windows-1251")

                if(decoded_date):
                    return email.utils.parsedate_to_datetime(decoded_date)

class BodyParser:
    def __init__(self, msg):
        self.msg = msg
        self.fails = {}

    def _parse_body_rec(self, msg):
        if msg.get_content_disposition() == "attachment":  # therefore attachment not body part
            return ""

        body = []
        message_content_type = msg.get_content_type()
        message_main_type = msg.get_content_maintype()

        if message_content_type == "multipart/mixed":
            for payload in msg.get_payload():
                body.append(self._parse_body_rec(payload))
        elif message_main_type == "text":
            body.append(self._parse_text_content(msg))
        elif message_content_type == "multipart/alternative":
            for payload in msg.get_payload():
                body.append(self._parse_body_rec(payload))
                break
        return "\n".join(body).strip()

    def parse_body(self):
        try:
            return self._parse_body_rec(self.msg)
        except:
            exc_type, value, _ = sys.exc_info()
            self.fails["body"] = "{}: {}".format(exc_type.__name__, value)
            return ""

    def _parse_text_content(self, msg):
        body = ""
        subtype = msg.get_content_subtype()
        charset = msg.get_content_charset(failobj="us-ascii")

        if subtype == "html":
            cleaned_body = self._clean_html(msg.get_payload(decode=True).decode(charset))
            body = cleaned_body
        elif subtype == "plain":
            if(msg["Content-Transfer-Encoding"] == "8bit"):
                try:
                    body = msg.get_payload(decode=True).decode(charset)
                except UnicodeDecodeError:
                    body = msg.get_payload(decode=True).decode("latin1")
            else:
                 body = msg.get_payload(decode=True).decode(charset)

        else:
            logging.warning("Content of this type not modified: {}".format(subtype))
            body = msg.get_payload(decode=True).decode(charset)

        return body

    def _parse_attachments_rec(self, msg, save_path, msg_id, max_attachment_size, random_filenames):
        attachments_jsons = []
        for payload in msg.get_payload():
            if type(payload) != str:
                if (payload.get_content_disposition() == "attachment"
                    and payload.get_content_type() != "message/rfc822"):

                    a_parser = AttachmentParser(payload, save_path, msg_id, max_attachment_size, random_filenames)
                    attachment_data = a_parser.parse()

                    if a_parser.fails != {}:
                        attachment_data["errors"] =  a_parser.fails

                    attachments_jsons.append(attachment_data)
                elif payload.is_multipart() and payload.get_content_type() != "message/rfc822":
                    attachments_jsons += self._parse_attachments_rec(payload, save_path, msg_id, max_attachment_size, random_filenames)

        return attachments_jsons

    def parse_attachments(self, save_path, msg_id, max_attachment_size, random_filenames):
        return self._parse_attachments_rec(self.msg, save_path, msg_id, max_attachment_size, random_filenames)

    def _clean_html(self, payload):
        soup = BeautifulSoup(payload, features="lxml")
        for script in soup(["script", "style"]):
            script.extract()
        text = soup.get_text()
        lines = (line.strip() for line in text.splitlines())
        text = "\n".join(line for line in lines if line)
        return text


class AttachmentParser(BodyParser):
    def __init__(self, msg, save_path, msg_id, max_attachment_size, random_filenames):
        BodyParser.__init__(self, msg)
        self.save_path = save_path
        self.msg_id = msg_id
        self.max_attachment_size = max_attachment_size
        self.random_filenames = random_filenames

    def parse(self):
        attachment_id = str(getrandbits(64))
        attachment_json = {"msg_id": self.msg_id, "attachment_id": attachment_id}

        filename = self._get_filename()
        ext = os.path.splitext(filename)[-1]
        attachment_json["filename"] = filename

        if(ext not in DIGIDOC_FORMATS):
            attachment_json["content"] = self._parse_attachment()

        if self.save_path != None and filename != "":
            attachment_dir = os.path.join(self.save_path, self.msg_id)

            if(self.random_filenames):
                ext = os.path.splitext(filename)[-1]
                random_filename = uuid.uuid4().hex + ext
                attachment_path = os.path.join(attachment_dir, random_filename)
            else:
                attachment_path = os.path.join(attachment_dir, filename)
            self._make_dir(attachment_dir)

            try:
                with open(attachment_path, "wb") as f:
                    f.write(self.msg.get_payload(decode=True))
                attachment_json["filepath"] = attachment_path  
            except OSError:
                exc_type, value, _ = sys.exc_info()
                self.fails["filename"] = "{}: {}".format(exc_type.__name__, value)
            except TypeError:
                exc_type, value, _ = sys.exc_info()
                self.fails["filepath"] = "{}: {}".format(exc_type.__name__, value)
                

        return attachment_json

    def _get_filename(self):
        try:
            encoded_filename = self.msg.get_filename()
            return AttachmentParser.parse_filename(encoded_filename)
        except:
            exc_type, value, _ = sys.exc_info()
            self.fails["filename"] = "{}: {}".format(exc_type.__name__, value)
            return ""

    @staticmethod
    def parse_filename(encoded_filename):
        filename_parts = email.header.decode_header(encoded_filename)

        filename_items = []
        for item_dec, charset in filename_parts:
            if charset:
                item_str = item_dec.decode(charset)
            elif type(item_dec) == bytes:
                item_str = item_dec.decode()
            else:
                assert type(item_dec) == str
                item_str = item_dec

            filename_items.append(item_str.strip())

        return "".join(filename_items)

    def _parse_attachment(self):
        attachment_payload = self.msg.get_payload(decode=True)
        if(attachment_payload == None):
            self.fails["content"] = "{}".format("No payload")
            return ""
        
        if(not self._file_size_ok(attachment_payload)):
            self.fails["content"] = "File too large for content parsing."
            return ""
        else:
            result = tika_parser.from_buffer(attachment_payload)
            content = result["content"]

            if result["status"] != 200:
                self.fails["content"] = "{}: {}".format("tika", result["status"])
            if content == None:
                return ""

            lines = (line.strip() for line in content.splitlines())
            text = "\n".join(line for line in lines if line)
            return text

    def _file_size_ok(self, payload):
        if(len(payload) > self.max_attachment_size*1024**2):
            return False
        return True

    def _make_dir(self, f_dir):
        if not os.path.exists(f_dir):
            os.makedirs(f_dir)



import csv
import os
import shutil
import uuid
from subprocess import Popen

import pandas as pd
from pyunpack import Archive
from tika import parser

from texta_parsers.email_parser import EmailParser
from texta_parsers.settings import META_FIELD
from texta_parsers.tools.extension import Extension
from texta_parsers.tools.scanner import DocScanner
from . import exceptions


# for digidoc parser
FNULL = open(os.devnull, "w")


class DocParser:

    def __init__(self, temp_dir="", languages=["est", "eng", "rus"]):
        """
        :param: list languages: Languages for TikaOCR.
        """
        self.temp_dir = self.create_temp_dir_for_parse(temp_dir)
        self.langs = languages
        self.scanner = DocScanner()


    @staticmethod
    def create_temp_dir_for_parse(temp_dir):
        """
        Creates temp directory path.
        """
        temp_dir_for_parse = os.path.join(temp_dir, "temp_" + uuid.uuid4().hex)
        if not os.path.exists(temp_dir_for_parse):
            os.mkdir(temp_dir_for_parse)
        return temp_dir_for_parse


    def _write_uploaded_to_file(self, uploadedfile, file_name):
        if not file_name:
            raise exceptions.InvalidInputError("File name not supported.")
        # get extension from file name if any
        extension = Extension.predict(uploadedfile, file_name=file_name)
        # create new path with predicted extension
        new_name = uuid.uuid4().hex + extension
        # new_name = uploadedfile.name
        file_path = os.path.join(self.temp_dir, new_name)
        with open(file_path, "wb") as fh:
            fh.write(uploadedfile)
        return file_path


    def _extract_digidoc(self, input_path, output_dir, extracted=[]):
        """
        Extracts contents from digidoc. Works recursively.
        """
        cmd = f"digidoc-tool open {input_path} --extractAll={output_dir}"
        p = Popen(cmd, shell=True, stdout=FNULL)
        p.wait()
        # generate full paths for the output
        extracted_docs = os.listdir(output_dir)
        extracted_docs = [os.path.join(output_dir, file_name) for file_name in extracted_docs]
        # extract further if digidocs in output
        for extracted_doc in extracted_docs:
            ext = Extension.predict(extracted_doc)
            if ext in Extension.DIGIDOC_EXTENSIONS:
                self._extract_digidoc(extracted_doc, output_dir, extracted=extracted)
            else:
                extracted.append(extracted_doc)
        return extracted


    def _extract_archive(self, input_path, output_dir, extension):
        """
        Extracts contents from archives. Works recursively.
        """
        # create temporary output directory
        # just to be safe, because file names may repeat
        output_dir = os.path.join(output_dir, uuid.uuid4().hex)
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)
        # extract with digidoc client
        if extension in Extension.DIGIDOC_EXTENSIONS:
            extracted_paths = self._extract_digidoc(input_path, output_dir)
        else:
            archive = Archive(input_path)
            archive.extractall(output_dir)
        # generate full paths for the output
        extracted_paths = os.listdir(output_dir)
        extracted_paths = [os.path.join(output_dir, file_name) for file_name in extracted_paths]
        # extract further if archives in output
        for extracted_path in extracted_paths:
            if os.path.isdir(extracted_path):
                ### TODO: HOLY SHIT ITS A DIRECTORY INSIDE AN ARCHIVE!
                pass
            else:
                extracted_extension = Extension.predict(extracted_path)
                if extracted_extension in Extension.ARCHIVE_EXTENSIONS:
                    for doc in self._extract_archive(extracted_path, output_dir, extracted_extension):
                        yield doc
                elif extracted_extension in Extension.KNOWN_EXTENSIONS:
                    yield {"path": extracted_path, "extension": extracted_extension}


    def remove_temp_dir(self):
        """
        Removes temp directory path.
        """
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)


    @staticmethod
    def _parse_file(document, languages):
        """
        Parses document using TIKA and Speech2Text
        """
        output_documents = []
        # TODO: WTF HAPPENS WITH OCR?
        tika_output = parser.from_file(document["path"], requestOptions={"headers": {"X-Tika-OCRLanguage": "+".join(languages)}})
        content = tika_output["content"]
        if content != None:  # remove leading and trailing spacing
            content = content.strip()

        output_documents.append({"text": content})
        return output_documents


    @staticmethod
    def _parse_collection(document):
        if (document["extension"] == ".csv"):
            # detect dialect and whether contains header
            with open(document["path"]) as f:
                lines = f.readline() + '\n' + f.readline()
                dialect = csv.Sniffer().sniff(lines)
                f.seek(0)

                has_header = csv.Sniffer().has_header(lines)
                f.seek(0)

            header = "infer" if has_header else None
            # read and yield actual data with pandas (more convenient)
            reader = pd.read_csv(document["path"], dialect=dialect, chunksize=1000, header=header)

            for chunk in reader:
                for ix, row in chunk.iterrows():
                    yield row.to_dict()

        # .xls or .xlsx
        else:
            reader = pd.read_excel(document["path"], header=0, sheet_name=None)  # dont know whether there is a header but assume that the first row is

            if isinstance(reader, dict):
                for sheet_name, df in reader.items():
                    df.fillna("", inplace=True)
                    for ix, row in df.iterrows():
                        yield row.to_dict()
            else:
                reader.fillna("", inplace=True)
                for ix, row in reader.iterrows():
                    yield row.to_dict()


    def _parse_mail(self, document):
        parser = EmailParser(tmp_folder=self.temp_dir)
        return parser.parse(document["path"])


    def parse(self, parser_input, file_name=None):
        """
        :param: str parser_input: Base64 string or file path.
        """
        if isinstance(parser_input, bytes):
            # input is in bytes
            file_paths = [self._write_uploaded_to_file(parser_input, file_name)]
        elif isinstance(parser_input, str):
            # input is path to file as string
            if not os.path.exists(parser_input):
                raise exceptions.InvalidInputError("File does not exist.")
            # input is a directory and we should scan it
            if os.path.isdir(parser_input):
                file_paths = self.scanner.scan(parser_input)
            else:
                file_paths = [parser_input]
        else:
            raise exceptions.InvalidInputError("Input should be path to file/directory or bytes.")

        # apply parsers for all paths in input
        for file_path in file_paths:
            print(file_path)
            docs_to_parse = []
            # guess extension (it also performs chck if extension is known)
            extension = Extension.predict(file_path, file_name=file_name)
            # check if an archive
            if extension in Extension.ARCHIVE_EXTENSIONS:
                docs_to_parse = self._extract_archive(file_path, self.temp_dir, extension)
            else:
                docs_to_parse.append({"path": file_path, "extension": extension})
            # parse file
            for meta in docs_to_parse:
                if (meta["extension"] in Extension.EMAIL_EXTENSIONS):
                    # returns email generator
                    gen = self._parse_mail(meta)
                    # yield items from it
                    for msg_dict, attachment_dicts in gen:
                        # add metadata to result
                        msg_dict[META_FIELD] = meta
                        for attachment in attachment_dicts:
                            attachment[META_FIELD] = meta
                        yield msg_dict, attachment_dicts
                elif (meta["extension"] in Extension.COLLECTION_EXTENSIONS):
                    # returns collection generator
                    gen = self._parse_collection(meta)
                    # yield items from it
                    for item in gen:
                        item[META_FIELD] = meta
                        yield item
                else:
                    for parsed_document in self._parse_file(meta, languages=self.langs):
                        parsed_document[META_FIELD] = meta
                        yield parsed_document



import os

from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname), encoding="utf8").read()


def get_requirements(fname):
    excluded_packages = ["pytest"]
    req_string = read(fname)
    return [req.strip() for req in req_string.strip().split("\n") if req.strip() not in excluded_packages]


setup(
    name="texta-parsers",
    version=read("VERSION"),
    author="TEXTA",
    author_email="info@texta.ee",
    description=("texta-parsers"),
    license="GPLv3",
    packages=['texta_parsers', 'texta_parsers/tools'],
    data_files=["VERSION", "requirements.txt", "README.md"],
    long_description_content_type="text/markdown",
    long_description=read('README.md'),
    install_requires=get_requirements("requirements.txt"),
)



import logging
import os
import warnings

import elasticsearch
from elasticsearch.helpers import bulk
from elasticsearch_dsl import Keyword, Long, Mapping, Nested

from . import utils
from .utils import ParserOutputType
from ..settings import META_FIELD


logging.basicConfig(
    format='%(levelname)s %(asctime)s: %(message)s',
    datefmt='%d.%m.%Y %H:%M:%S',
    level=logging.INFO
)


class ESImporter:


    def __init__(self, elastic_url: str, index_prefix: str):
        """
        :param elastic_url: Url to elasticsearch
        :param index_prefix: Prefix used for naming indices in elasticsearch
        """
        self.index_prefix = index_prefix
        self.es = elasticsearch.Elasticsearch(elastic_url)


    def _importer_generator(self, generator, index_name: str):
        """
        Used to generate the payload for the Bulk API of the Python Elasticsearch Clients
        helper.
        :param generator: Output of the entity-linker (concatenator)
        :param index_name: Full name of the index into which the document is pushed.
        """
        for item in generator:
            yield {
                "_index": index_name,
                "_type": index_name,
                "_source": item
            }


    def push_linked_entities_into_elastic(self, generator, index_name: str) -> bool:
        """
        Generator for pushing the EntityLinker results into an Elasticsearch index.

        :param generator: Generator which yields the results of the entity-linking (concatenation)
        :param index_name: Full name of the index you want to push the results into.
        :return: As this is meant as the final part of the pipeline, this will only return True.
        """
        self._add_texta_facts_mapping(index=index_name)
        generator = self._importer_generator(generator, index_name)
        generator = list(generator)
        bulk(client=self.es, actions=generator, refresh="wait_for")
        logging.info("Finished pushing person info into Elasticsearch!")
        return True


    def insert_from_generator(self, generator, toolkit_url=None, tk_media_folder=None):
        """ Inserts data to elasticsearch from given generator object.
           
        # Parameters
        toolkit_url(str): ONLY FOR EMAILS! If this is specified then filepath field is converted to an HTML link by combining toolkit URL with filepath. This is only done
        for convenient purpose to make it possible to wiew and download original files directly from toolkit. However, assumes that original files are saved to toolkit's
        media folder during parsing, otherwise the links won't work.
        tk_media_folder(str): ONLY FOR EMAILS! A path to toolkit's media folder. Is only used when toolkit_url is set.
        """
        if (toolkit_url != None and tk_media_folder == None):
            raise ValueError("Argument tk_media_folder can not be None if toolkit_url is given.")

        bulk(actions=self._get_es_data(generator, toolkit_url, tk_media_folder), client=self.es, refresh="wait_for")


    def _get_es_data(self, generator, toolkit_url, tk_media_folder):
        has_mails_mapping = False
        has_attachments_mapping = False
        has_index_mapping = False

        for item in generator:

            # Get index name. It depends on the original file extension
            index_name = self._get_index_name(item)

            # Check if from email generator
            if type(item) == tuple:
                # Append "mails" or "attachments" to index name to put them to different indices
                mails_index = index_name + "_mails"
                attachments_index = index_name + "_attachments"
                mail, attachments = item

                self._try_add_file_url(mail, toolkit_url, tk_media_folder)
                if has_mails_mapping is False:
                    self._add_texta_facts_mapping(mails_index)
                    has_mails_mapping = True

                yield {
                    "_index": mails_index,
                    "_type": mails_index,
                    "_source": mail
                }

                for attachment in attachments:
                    if has_attachments_mapping is False:
                        self._add_texta_facts_mapping(attachments_index)
                        has_attachments_mapping = True

                    self._try_add_file_url(attachment, toolkit_url, tk_media_folder)
                    yield {
                        "_index": attachments_index,
                        "_type": attachments_index,
                        "_source": attachment
                    }
            else:

                if has_index_mapping is False:
                    self._add_texta_facts_mapping(index_name)
                    has_index_mapping = True

                yield {
                    "_index": index_name,
                    "_type": index_name,
                    "_source": item
                }


    def _add_texta_facts_mapping(self, index):
        self.es.indices.create(index, ignore=[400, 404])

        m = Mapping(index)
        texta_facts = Nested(
            properties={
                "spans": Keyword(),
                "fact": Keyword(),
                "str_val": Keyword(),
                "doc_path": Keyword(),
                "num_val": Long(),
            }
        )

        # Set the name of the field along with its mapping body
        m.field("texta_facts", texta_facts)
        m.save(index=index, using=self.es)


    def _get_index_name(self, item):
        item_type = utils.get_output_type(item)

        if (item_type == ParserOutputType.EMAIL):
            # just return prefix, finish naming later
            return self.index_prefix
        elif (item_type == ParserOutputType.COLLECTION):
            # use file name in naming as well to put different collections to different indices
            filename = os.path.basename(item[META_FIELD]["path"]).replace(".", "_").lower()
            return self.index_prefix + "_collection_" + filename
        else:
            # put everything else (texts, images, audio etc.) to general "files" index
            return self.index_prefix + "_files"


    def _try_add_file_url(self, item, toolkit_url, tk_media_folder):
        if item != {}:
            if toolkit_url != None:
                tk_url_path = self._get_file_url(item, toolkit_url, tk_media_folder)
                if (tk_url_path != None):
                    item["filepath"] = tk_url_path


    def _get_file_url(self, data_json, toolkit_url, tk_media_folder):
        if "filepath" in data_json:
            data_filepath = data_json["filepath"]
            if (data_filepath.startswith(tk_media_folder) == False):
                warnings.warn("File with filepath '{0}' is not located in '{1}'. Can not create URL for this file.".format(data_filepath, tk_media_folder))
            else:
                file_part = data_filepath.replace(tk_media_folder, "").strip("/")
                file_url = os.path.join(toolkit_url, "data/media", file_part)
                return file_url



class ServiceNotAvailableError(Exception):
    """Raised when external service is not available.""" 
    pass

class MLPFailedError(Exception):
    """Raised when MLP processing fails."""
    pass

class InvalidInputError(Exception):
    """Raised when something incorrect given to trainers."""
    pass

class UnsupportedFileError(Exception):
    """Raised when unsupported file given to parser."""
    pass




from enum import Enum
from ..settings import META_FIELD
from texta_parsers.tools.extension import Extension


class ParserOutputType(Enum):
     EMAIL = 1
     COLLECTION = 2
     FILE = 3

def get_output_type(item):
    #item is from email generator in the form of (email_dict, [attachment_dict])
    #can not check extension here because input can come directly from email parser as well
    #which does not have that field
    if(type(item) == tuple):
        return ParserOutputType.EMAIL
    #item is from docparser generator, find type by checking extension
    else:
        extension = item[META_FIELD]["extension"]
        if(extension in Extension.COLLECTION_EXTENSIONS):
            return ParserOutputType.COLLECTION
        else:
            return ParserOutputType.FILE
